@@ -1,6 +1,6 @@
 package Sub::Spec::CmdLine;
 BEGIN {
-  $Sub::Spec::CmdLine::VERSION = '0.17';
+  $Sub::Spec::CmdLine::VERSION = '0.18';
 }
 # ABSTRACT: Access Perl subs via command line
 
@@ -57,11 +57,11 @@ sub parse_argv {
             #$go_spec{$opt} = sub { $args->{$name[0]} = $_[0] };
             $go_spec{$opt} = \$args->{$name[0]};
         }
-        my $aliases = $schema->{attr_hashes}[0]{cmdline_aliases};
+        my $aliases = $schema->{attr_hashes}[0]{arg_aliases};
         if ($aliases) {
             while (my ($alias, $alinfo) = each %$aliases) {
                 my $opt;
-                if ($schema->{type} eq 'bool' && !$alinfo->{code}) {
+                if ($schema->{type} eq 'bool') {
                     $opt = "$alias!";
                 } else {
                     $opt = "$alias=s";
@@ -239,7 +239,7 @@ sub gen_usage($;$) {
             Data::Dump::Partial::dumpp($ah0->{default}).")"
                   if defined($ah0->{default});
 
-        my $aliases = $ah0->{cmdline_aliases};
+        my $aliases = $ah0->{arg_aliases};
         if ($aliases) {
             $arg_desc .= "\n";
             for (sort keys %$aliases) {
@@ -249,7 +249,7 @@ sub gen_usage($;$) {
                     "      ",
                     (length == 1 ? "-$_" : "--$_"), " ",
                     $alinfo->{summary} ? $alinfo->{summary} :
-                        "is alias for --$name",
+                        "is alias for '$name'",
                     "\n"
                 );
             }
@@ -382,18 +382,72 @@ sub _run_completion {
         for (@o) { push @general_opts, length > 1 ? "--$_" : "-$_" }
     }
 
-    my $spec = $args{spec};
-    if ($spec && ($args{space_typed} || !$args{subcommand})) {
+    my $spec  = $args{spec};
+    my $subc  = $args{subcommand};
+    my $subcn = $args{subcommand_name};
+    my $words = $args{words};
+    my $cword = $args{cword};
+
+    # whether we should complete arg names/values or general opts + subcommands
+    # name
+    my $do_arg;
+    {
+        # we can't do arg unless we already get the spec
+        if (!$spec) {
+            $log->trace("not do_arg because there is no spec");
+            last;
+        }
+
+        # single-sub directly complete arg names/values
+        if (!$subc) {
+            $log->trace("do_arg because single sub");
+            $do_arg++; last;
+        }
+
+        # multiple-sub, just typing "CMD subc ^" (space already typed)
+        if ($cword > 0 && $args{space_typed} && $subcn) {
+            $log->trace("do_arg because last word typed (+space) is ".
+                            "subcommand name");
+            $do_arg++; last;
+        }
+
+        # multiple-sub, already typing subc in past words
+        if ($cword > 0 && !$args{space_typed} && $words->[$cword] ne $subcn) {
+            $log->trace("do_arg because subcommand name has been typed ".
+                            "in past words");
+            $do_arg++; last;
+        }
+
+        $log->tracef("not do_arg, cword=%d, words=%s, subcommand_name=%s, ".
+                         "space_typed=%s",
+                     $cword, $words, $subcn, $args{space_typed});
+    }
+    if ($do_arg) {
         $log->trace("Complete subcommand argument names & values");
+
+        # remove subcommand name and general options from words so it doesn't
+        # interfere with matching spec args
+        my $i = 0;
+        while ($i < @$words) {
+            if ($words->[$i] ~~ @general_opts || $words->[$i] eq $subcn) {
+                splice @$words, $i, 1;
+                $cword-- unless $cword <= $i;
+                next;
+            } else {
+                $i++;
+            }
+        }
+        $log->tracef("cleaned words=%s, cword=%d", $words, $cword);
+
         return Sub::Spec::BashComplete::bash_complete_spec_arg(
             $spec,
             {
-                words            => $args{words},
-                cword            => $args{cword},
+                words            => $words,
+                cword            => $cword,
                 arg_sub          => $args{arg_sub},
                 args_sub         => $args{args_sub},
                 custom_completer =>
-                    ($args{subcommand} ? $args{subcommand}{custom_completer} :
+                    ($subc ? $subc->{custom_completer} :
                          undef) // $args{parent_args}{custom_completer}
             },
         );
@@ -451,6 +505,7 @@ _
 }
 
 sub run {
+    $log->trace("-> run()");
     require Getopt::Long;
 
     my %args = @_;
@@ -563,9 +618,6 @@ sub run {
         my $complete_arg;
         my $complete_args;
         if ($subc) {
-            shift @$comp_words;
-            $comp_cword-- unless $comp_cword < 1;
-
             $complete_arg    = $subc->{complete_arg};
             $complete_args   = $subc->{complete_args};
         }
@@ -634,8 +686,11 @@ sub run {
     if ($subc && $subc->{run}) {
         # use run routine instead if supplied
         $res = $subc->{run}->(
-            module=>$module, sub=>$sub, spec=>$spec,
-            args=>$args,
+            subcommand_name => $subc_name,
+            module   => $module,
+            sub      => $sub,
+            spec     => $spec,
+            sub_args => $args,
         );
     } else {
         # call sub
@@ -663,7 +718,7 @@ Sub::Spec::CmdLine - Access Perl subs via command line
 
 =head1 VERSION
 
-version 0.17
+version 0.18
 
 =head1 SYNOPSIS
 
